#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <openssl/sha.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/mount.h>
#include <sys/xattr.h>
#include <sys/types.h>
#include <zlib.h>

#include <CoreServices/CoreServices.h>

void compressFile(char *inFile, struct stat *inFileInfo, long long int maxSize, int compressionlevel, int minSavings, bool checkFiles)
{
	FILE *in;
	struct statfs fsInfo;
	unsigned int compblksize = 0x10000, numBlocks, outdecmpfsSize = 0;
	void *inBuf, *outBuf, *outBufBlock, *outdecmpfsBuf, *currBlock, *blockStart;
	long long int inBufPos, filesize = inFileInfo->st_size;
	unsigned long int cmpedsize;
	char *xattrnames, *curr_attr;
	ssize_t xattrnamesize;
	UInt32 cmpf = 0x636D7066;
	struct timeval times[2];
	
	times[0].tv_sec = inFileInfo->st_atimespec.tv_sec;
	times[0].tv_usec = inFileInfo->st_atimespec.tv_nsec / 1000;
	times[1].tv_sec = inFileInfo->st_mtimespec.tv_sec;
	times[1].tv_usec = inFileInfo->st_mtimespec.tv_nsec / 1000;
	
	if (statfs(inFile, &fsInfo) < 0)
		return;
	if (fsInfo.f_type != 17)
		return;
	
	if (!S_ISREG(inFileInfo->st_mode))
		return;
	if ((inFileInfo->st_flags & UF_COMPRESSED) != 0)
		return;
	if (filesize > maxSize)
		return;
	if (filesize == 0)
		return;
	
	if (chflags(inFile, UF_COMPRESSED | inFileInfo->st_flags) < 0 || chflags(inFile, inFileInfo->st_flags) < 0)
		return;
	
	xattrnamesize = listxattr(inFile, NULL, 0, XATTR_SHOWCOMPRESSION | XATTR_NOFOLLOW);
	
	if (xattrnamesize > 0)
	{
		xattrnames = (char *) malloc(xattrnamesize);
		if (xattrnames == NULL)
		{
			fprintf(stderr, "%s: malloc error, unable to get file information\n", inFile);
			return;
		}
		if ((xattrnamesize = listxattr(inFile, xattrnames, xattrnamesize, XATTR_SHOWCOMPRESSION | XATTR_NOFOLLOW)) <= 0)
		{
			fprintf(stderr, "%s: listxattr: %s\n", inFile, strerror(errno));
			free(xattrnames);
			return;
		}
		for (curr_attr = xattrnames; curr_attr < xattrnames + xattrnamesize; curr_attr += strlen(curr_attr) + 1)
		{
			if ((strcmp(curr_attr, "com.apple.ResourceFork") == 0 && strlen(curr_attr) == 22) ||
				(strcmp(curr_attr, "com.apple.decmpfs") == 0 && strlen(curr_attr) == 17))
				return;
		}
		free(xattrnames);
	}
	
	numBlocks = (filesize + compblksize - 1) / compblksize;
	
	in = fopen(inFile, "r+");
	if (in == NULL)
	{
		fprintf(stderr, "%s: %s\n", inFile, strerror(errno));
		return;
	}
	inBuf = malloc(filesize);
	if (inBuf == NULL)
	{
		fprintf(stderr, "%s: malloc error, unable to allocate input buffer\n", inFile);
		fclose(in);
		utimes(inFile, times);
		return;
	}
	if (fread(inBuf, filesize, 1, in) != 1)
	{
		fprintf(stderr, "%s: Error reading file\n", inFile);
		fclose(in);
		utimes(inFile, times);
		free(inBuf);
		return;
	}
	fclose(in);
	outBuf = malloc(filesize + 0x13A + (numBlocks * 8));
	if (outBuf == NULL)
	{
		fprintf(stderr, "%s: malloc error, unable to allocate output buffer\n", inFile);
		utimes(inFile, times);
		free(inBuf);
		return;
	}
	outdecmpfsBuf = malloc(3802);
	if (outdecmpfsBuf == NULL)
	{
		fprintf(stderr, "%s: malloc error, unable to allocate xattr buffer\n", inFile);
		utimes(inFile, times);
		free(inBuf);
		free(outBuf);
		return;
	}
	outBufBlock = malloc(compressBound(compblksize));
	if (outBufBlock == NULL)
	{
		fprintf(stderr, "%s: malloc error, unable to allocate compression buffer\n", inFile);
		utimes(inFile, times);
		free(inBuf);
		free(outBuf);
		free(outdecmpfsBuf);
		return;
	}
	*(UInt32 *) outdecmpfsBuf = EndianU32_NtoL(cmpf);
	*(UInt32 *) (outdecmpfsBuf + 4) = EndianU32_NtoL(4);
	*(UInt64 *) (outdecmpfsBuf + 8) = EndianU64_NtoL(filesize);
	outdecmpfsSize = 0x10;
	*(UInt32 *) outBuf = EndianU32_NtoB(0x100);
	*(UInt32 *) (outBuf + 12) = EndianU32_NtoB(0x32);
	memset(outBuf + 16, 0, 0xF0);
	blockStart = outBuf + 0x104;
	*(UInt32 *) blockStart = EndianU32_NtoL(numBlocks);
	currBlock = blockStart + 0x4 + (numBlocks * 8);
	for (inBufPos = 0; inBufPos < filesize; inBufPos += compblksize, currBlock += cmpedsize)
	{
		cmpedsize = compressBound(compblksize);
		if (compress2(outBufBlock, &cmpedsize, inBuf + inBufPos, ((filesize - inBufPos) > compblksize) ? compblksize : filesize - inBufPos, compressionlevel) != Z_OK)
		{
			utimes(inFile, times);
			free(inBuf);
			free(outBuf);
			free(outdecmpfsBuf);
			free(outBufBlock);
			return;
		}
		if (((cmpedsize + outdecmpfsSize) <= 3802) && (numBlocks <= 1))
		{
			*(UInt32 *) (outdecmpfsBuf + 4) = EndianU32_NtoL(3);
			memcpy(outdecmpfsBuf + outdecmpfsSize, outBufBlock, cmpedsize);
			outdecmpfsSize += cmpedsize;
			break;
		}
		if (cmpedsize > (((filesize - inBufPos) > compblksize) ? compblksize : filesize - inBufPos))
		{
			utimes(inFile, times);
			free(inBuf);
			free(outBuf);
			free(outdecmpfsBuf);
			free(outBufBlock);
			return;
		}
		memcpy(currBlock, outBufBlock, cmpedsize);
		*(UInt32 *) (blockStart + ((inBufPos / compblksize) * 8) + 0x4) = EndianU32_NtoL(currBlock - blockStart);
		*(UInt32 *) (blockStart + ((inBufPos / compblksize) * 8) + 0x8) = EndianU32_NtoL(cmpedsize);
	}
	
	if (EndianU32_LtoN(*(UInt32 *) (outdecmpfsBuf + 4)) == 4)
	{
		if (((double) (currBlock - outBuf) / filesize) >= (1.0 - ((float) minSavings / 100)))
		{
			utimes(inFile, times);
			free(inBuf);
			free(outBuf);
			free(outdecmpfsBuf);
			free(outBufBlock);
			return;
		}
		*(UInt32 *) (outBuf + 4) = EndianU32_NtoB(currBlock - outBuf);
		*(UInt32 *) (outBuf + 8) = EndianU32_NtoB(currBlock - outBuf - 0x100);
		*(UInt32 *) (blockStart - 4) = EndianU32_NtoB(currBlock - outBuf - 0x104);
		memset(currBlock, 0, 24);
		*(UInt16 *) (currBlock + 24) = EndianU16_NtoB(0x1C);
		*(UInt16 *) (currBlock + 26) = EndianU16_NtoB(0x32);
		*(UInt16 *) (currBlock + 28) = 0;
		*(UInt32 *) (currBlock + 30) = EndianU32_NtoB(cmpf);
		*(UInt32 *) (currBlock + 34) = EndianU32_NtoB(0xA);
		*(UInt64 *) (currBlock + 38) = EndianU64_NtoL(0xFFFF0100);
		*(UInt32 *) (currBlock + 46) = 0;
		if (setxattr(inFile, "com.apple.ResourceFork", outBuf, currBlock - outBuf + 50, 0, XATTR_NOFOLLOW | XATTR_CREATE) < 0)
		{
			fprintf(stderr, "%s: setxattr: %s\n", inFile, strerror(errno));
			free(inBuf);
			free(outBuf);
			free(outdecmpfsBuf);
			free(outBufBlock);
			return;
		}
	}
	if (setxattr(inFile, "com.apple.decmpfs", outdecmpfsBuf, outdecmpfsSize, 0, XATTR_NOFOLLOW | XATTR_CREATE) < 0)
	{
		fprintf(stderr, "%s: setxattr: %s\n", inFile, strerror(errno));
		free(inBuf);
		free(outBuf);
		free(outdecmpfsBuf);
		free(outBufBlock);
		return;
	}
	in = fopen(inFile, "w");
	if (in == NULL)
	{
		fprintf(stderr, "%s: %s\n", inFile, strerror(errno));
		return;
	}
	fclose(in);
	if (chflags(inFile, UF_COMPRESSED | inFileInfo->st_flags) < 0)
	{
		fprintf(stderr, "%s: chflags: %s\n", inFile, strerror(errno));
		if (removexattr(inFile, "com.apple.decmpfs", XATTR_NOFOLLOW | XATTR_SHOWCOMPRESSION) < 0)
		{
			fprintf(stderr, "%s: removexattr: %s\n", inFile, strerror(errno));
		}
		if (EndianU32_LtoN(*(UInt32 *) (outdecmpfsBuf + 4)) == 4 &&
			removexattr(inFile, "com.apple.ResourceFork", XATTR_NOFOLLOW | XATTR_SHOWCOMPRESSION) < 0)
		{
			fprintf(stderr, "%s: removexattr: %s\n", inFile, strerror(errno));
		}
		in = fopen(inFile, "w");
		if (in == NULL)
		{
			free(inBuf);
			free(outBuf);
			free(outdecmpfsBuf);
			free(outBufBlock);
			fprintf(stderr, "%s: %s\n", inFile, strerror(errno));
			return;
		}
		if (fwrite(inBuf, filesize, 1, in) != 1)
		{
			free(inBuf);
			free(outBuf);
			free(outdecmpfsBuf);
			free(outBufBlock);
			fprintf(stderr, "%s: Error writing to file\n", inFile);
			return;
		}
		fclose(in);
		utimes(inFile, times);
		return;
	}
	if (checkFiles)
	{
		lstat(inFile, inFileInfo);
		in = fopen(inFile, "r");
		if (in == NULL)
		{
			fprintf(stderr, "%s: %s\n", inFile, strerror(errno));
			return;
		}
		if (inFileInfo->st_size != filesize || 
			fread(outBuf, filesize, 1, in) != 1 ||
			memcmp(outBuf, inBuf, filesize) != 0)
		{
			fclose(in);
			printf("%s: Compressed file check failed, reverting file changes\n", inFile);
			if (chflags(inFile, (~UF_COMPRESSED) & inFileInfo->st_flags) < 0)
			{
				free(inBuf);
				free(outBuf);
				free(outdecmpfsBuf);
				free(outBufBlock);
				fprintf(stderr, "%s: chflags: %s\n", inFile, strerror(errno));
				return;
			}
			if (removexattr(inFile, "com.apple.decmpfs", XATTR_NOFOLLOW | XATTR_SHOWCOMPRESSION) < 0)
			{
				fprintf(stderr, "%s: removexattr: %s\n", inFile, strerror(errno));
			}
			if (EndianU32_LtoN(*(UInt32 *) (outdecmpfsBuf + 4)) == 4 && 
				removexattr(inFile, "com.apple.ResourceFork", XATTR_NOFOLLOW | XATTR_SHOWCOMPRESSION) < 0)
			{
				fprintf(stderr, "%s: removexattr: %s\n", inFile, strerror(errno));
			}
			in = fopen(inFile, "w");
			if (in == NULL)
			{
				free(inBuf);
				free(outBuf);
				free(outdecmpfsBuf);
				free(outBufBlock);
				fprintf(stderr, "%s: %s\n", inFile, strerror(errno));
				return;
			}
			if (fwrite(inBuf, filesize, 1, in) != 1)
			{
				free(inBuf);
				free(outBuf);
				free(outdecmpfsBuf);
				free(outBufBlock);
				fprintf(stderr, "%s: Error writing to file\n", inFile);
				return;
			}
		}
		fclose(in);
	}
	utimes(inFile, times);
	free(inBuf);
	free(outBuf);
	free(outdecmpfsBuf);
	free(outBufBlock);
}

int main(int argc, char *argv[])
{
	FILE *foundationFramework;
	long fileLen;
	unsigned char *fileBuf, fileSHA1[20], SHA1str[41];
	int i;
	int currBase = 0;
	const char foundationExec[] = "/System/Library/Frameworks/Foundation.framework/Versions/C/Foundation";
	struct stat fileinfo;
	
	if (geteuid() != 0)
	{
		fprintf(stderr, "This must be run as root (try \"sudo %s\")\n", argv[0]);
		exit(EPERM);
	}

	if (lstat(foundationExec, &fileinfo) < 0)
	{
		fprintf(stderr, "%s: %s\n", argv[i], strerror(errno));
		return -1;
	}
	
	foundationFramework = fopen(foundationExec, "r");
	if (foundationFramework == NULL)
	{
		fprintf(stderr, "Error reading Foundation framework executable\n");
		exit(EPERM);
	}
	fseek(foundationFramework, 0, SEEK_END);
	fileLen = ftell(foundationFramework);
	rewind(foundationFramework);
	fileBuf = (unsigned char *) malloc(fileLen);
	if (fileBuf == NULL)
	{
		fprintf(stderr, "Malloc error\n");
		exit(ENOMEM);
	}
	if (fread(fileBuf, fileLen, 1, foundationFramework) != 1)
	{
		fprintf(stderr, "Error reading Foundation framework executable\n");
		exit(EIO);
	}
	fclose(foundationFramework);
	SHA1(fileBuf, fileLen, fileSHA1);
	
	sprintf(SHA1str, "%02x", fileSHA1[0]);
	for (i = 1; i < 20; i++)
		sprintf(SHA1str, "%s%02x", SHA1str, fileSHA1[i]);
	if (memcmp(SHA1str, "015ace4bdf63420e623951dd3ed93c25d5b72611", 40) == 0)
		currBase = 10;
	else if (memcmp(SHA1str, "cb3b9a5af8a9d7b3a17d06b1c21dcc7b51977207", 40) == 0)
		currBase = 2;
	
	if (currBase == 0)
	{
		fprintf(stderr, "Cannot patch Foundation framework; either it is not version 6.6 or it is corrupted\n");
		exit(EFTYPE);
	}
	
	foundationFramework = fopen(foundationExec, "r+");
	if (foundationFramework == NULL)
	{
		fprintf(stderr, "Error patching Foundation framework executable\n");
		exit(EPERM);
	}
	
	if (currBase == 10)
	{
		fseek(foundationFramework, 0x34147, SEEK_SET);
		putc(0x00, foundationFramework);
		fseek(foundationFramework, 0x43095C, SEEK_SET);
		putc(0x00, foundationFramework);
	}
	else if (currBase == 2)
	{
		fseek(foundationFramework, 0x34147, SEEK_SET);
		putc(0xE8, foundationFramework);
		fseek(foundationFramework, 0x43095C, SEEK_SET);
		putc(0xE8, foundationFramework);
	}
	
	fclose(foundationFramework);
	compressFile((char *) foundationExec, &fileinfo, 20971520, 5, 25, TRUE);
	
	printf("Successfully switched Foundation framework disk size calculations to base %d, restart your computer for the changes to completely take effect\n", (currBase == 10) ? 2 : 10);
	
	return 0;
}
